@@ -19,8 +19,7 @@ import grpc
 
 from e6data_python_connector.common import DBAPITypeObject, ParamEscaper, DBAPICursor
 from e6data_python_connector.constants import *
-from e6data_python_connector.datainputstream import DataInputStream, get_query_columns_info, read_rows_from_batch, \
-    read_values_from_array
+from e6data_python_connector.datainputstream import get_query_columns_info, read_rows_from_chunk
 from e6data_python_connector.server import e6x_engine_pb2_grpc, e6x_engine_pb2
 from e6data_python_connector.typeId import *
 
@@ -84,6 +83,10 @@ class HiveParamEscaper(ParamEscaper):
 _escaper = HiveParamEscaper()
 
 
+def _get_grpc_header(engine_ip):
+    return [('plannerip', engine_ip)]
+
+
 def connect(*args, **kwargs):
     """Constructor for creating a connection to the database. See class :py:class:`Connection` for
     arguments.
@@ -110,7 +113,8 @@ class Connection(object):
             ssl_cert=None,
             thrift_transport=None,
             grpc_options=None,
-            catalog=None
+            catalog=None,
+            secure_channel=False
     ):
         if not username or not password:
             raise ValueError("username or password cannot be empty.")
@@ -123,10 +127,13 @@ class Connection(object):
         self._host = host
         self._port = port
 
+        self._ssl_cert = ssl_cert
+        self._secure_channel = secure_channel
+
         self.catalog_name = catalog
 
         self._keepalive_timeout_ms = 900000
-        self._max_receive_message_length = 100 * 1024 * 1024  # mb
+        self._max_receive_message_length = -1
         self._max_send_message_length = 300 * 1024 * 1024  # mb
 
         if type(grpc_options) == dict:
@@ -137,14 +144,25 @@ class Connection(object):
         self._create_client()
 
     def _create_client(self):
-        self._channel = grpc.insecure_channel(
-            target='{}:{}'.format(self._host, self._port),
-            options=[
-                ("grpc.keepalive_timeout_ms", self._keepalive_timeout_ms),
-                ('grpc.max_send_message_length', self._max_send_message_length),
-                ('grpc.max_receive_message_length', self._max_receive_message_length)
-            ],
-        )
+        if self._secure_channel:
+            self._channel = grpc.secure_channel(
+                target='{}:{}'.format(self._host, self._port),
+                options=[
+                    ("grpc.keepalive_timeout_ms", self._keepalive_timeout_ms),
+                    ('grpc.max_send_message_length', self._max_send_message_length),
+                    ('grpc.max_receive_message_length', self._max_receive_message_length)
+                ],
+                credentials=grpc.ssl_channel_credentials()
+            )
+        else:
+            self._channel = grpc.insecure_channel(
+                target='{}:{}'.format(self._host, self._port),
+                options=[
+                    ("grpc.keepalive_timeout_ms", self._keepalive_timeout_ms),
+                    ('grpc.max_send_message_length', self._max_send_message_length),
+                    ('grpc.max_receive_message_length', self._max_receive_message_length)
+                ]
+            )
         self._client = e6x_engine_pb2_grpc.QueryEngineServiceStub(self._channel)
 
     @property
@@ -154,7 +172,6 @@ class Connection(object):
         """
         if not self._session_id:
             try:
-
                 authenticate_request = e6x_engine_pb2.AuthenticateRequest(
                     user=self.__username,
                     password=self.__password
@@ -201,7 +218,10 @@ class Connection(object):
             queryId=query_id,
             engineIP=engine_ip
         )
-        self._client.clear(clear_request)
+        self._client.clear(
+            clear_request,
+            metadata=_get_grpc_header(engine_ip)
+        )
         self._session_id = None
 
     def reopen(self):
@@ -214,7 +234,10 @@ class Connection(object):
             sessionId=self.get_session_id,
             queryId=query_id
         )
-        self._client.cancelQuery(cancel_query_request)
+        self._client.cancelQuery(
+            cancel_query_request,
+            metadata=_get_grpc_header(engine_ip)
+        )
 
     def dry_run(self, query):
         dry_run_request = e6x_engine_pb2.DryRunRequest(
@@ -274,7 +297,6 @@ class Cursor(DBAPICursor):
     Cursors are not isolated, i.e., any changes done to the database by a cursor are immediately
     visible by other cursors or connections.
     """
-    rows_count = 0
 
     def __init__(self, connection: Connection, array_size=1000, database=None, catalog_name=None):
         super(Cursor, self).__init__()
@@ -295,6 +317,10 @@ class Cursor(DBAPICursor):
         """Reset state about the previous query in preparation for running another query"""
         pass
 
+    @property
+    def metadata(self):
+        return _get_grpc_header(self._engine_ip)
+
     @property
     def arraysize(self):
         return self._arraysize
@@ -375,7 +401,7 @@ class Cursor(DBAPICursor):
             queryId=query_id,
             engineIP=self._engine_ip
         )
-        return self.connection.client.clear(clear_request)
+        return self.connection.client.clear(clear_request, metadata=self.metadata)
 
     def cancel(self, query_id):
         self.connection.query_cancel(engine_ip=self._engine_ip, query_id=query_id)
@@ -386,7 +412,7 @@ class Cursor(DBAPICursor):
             queryId=query_id,
             engineIP=self._engine_ip
         )
-        return self.connection.client.status(status_request)
+        return self.connection.client.status(status_request, metadata=self.metadata)
 
     def execute(self, operation, parameters=None, **kwargs):
         """Prepare and execute a database operation (query or command).
@@ -395,6 +421,7 @@ class Cursor(DBAPICursor):
         """
         Semicolon is now not supported. So removing it from query end.
         """
+        operation = operation.strip()  # Remove leading and trailing whitespaces.
         if operation.endswith(';'):
             operation = operation[:-1]
 
@@ -418,15 +445,15 @@ class Cursor(DBAPICursor):
             execute_statement_request = e6x_engine_pb2.ExecuteStatementRequest(
                 engineIP=self._engine_ip,
                 sessionId=self.connection.get_session_id,
-                queryId=self._query_id
+                queryId=self._query_id,
             )
             execute_statement_response = client.executeStatement(execute_statement_request)
         else:
             prepare_statement_request = e6x_engine_pb2.PrepareStatementV2Request(
                 sessionId=self.connection.get_session_id,
                 schema=self._database,
-                queryString=sql,
                 catalog=self._catalog_name,
+                queryString=sql
             )
             prepare_statement_response = client.prepareStatementV2(prepare_statement_request)
 
@@ -437,11 +464,16 @@ class Cursor(DBAPICursor):
                 sessionId=self.connection.get_session_id,
                 queryId=self._query_id
             )
-            execute_statement_response = client.executeStatementV2(execute_statement_request)
+            execute_statement_response = client.executeStatementV2(
+                execute_statement_request,
+                metadata=self.metadata
+            )
         self.update_mete_data()
         return self._query_id
 
+    @property
     def rowcount(self):
+        self.update_mete_data()
         return self._rowcount
 
     def update_mete_data(self):
@@ -450,9 +482,13 @@ class Cursor(DBAPICursor):
             sessionId=self.connection.get_session_id,
             queryId=self._query_id
         )
-        get_result_metadata_response = self.connection.client.getResultMetadata(result_meta_data_request)
+        get_result_metadata_response = self.connection.client.getResultMetadata(
+            result_meta_data_request,
+            metadata=self.metadata
+        )
         buffer = BytesIO(get_result_metadata_response.resultMetaData)
         self._rowcount, self._query_columns_description = get_query_columns_info(buffer)
+        self._is_metadata_updated = True
 
     def _fetch_more(self):
         batch_size = self._arraysize
@@ -485,37 +521,34 @@ class Cursor(DBAPICursor):
             yield rows
 
     def fetch_batch(self):
-        # _logger.debug("fetching next batch from e6data")
         client = self.connection.client
         get_next_result_batch_request = e6x_engine_pb2.GetNextResultBatchRequest(
             engineIP=self._engine_ip,
             sessionId=self.connection.get_session_id,
-            queryId=self._query_id,
-            asRowData=True
+            queryId=self._query_id
+        )
+        get_next_result_batch_response = client.getNextResultBatch(
+            get_next_result_batch_request,
+            metadata=self.metadata
         )
-        get_next_result_batch_response = client.getNextResultBatch(get_next_result_batch_request)
         buffer = get_next_result_batch_response.resultBatch
         if not self._is_metadata_updated:
             self.update_mete_data()
-            self._is_metadata_updated = True
-        if not buffer:
+        if not buffer or len(buffer) == 0:
             return None
-        buffer = BytesIO(buffer)
-        dis = DataInputStream(buffer)
         # one batch retrieves the predefined set of rows
-        return read_rows_from_batch(self._query_columns_description, dis)
+        return read_rows_from_chunk(self._query_columns_description, buffer)
 
     def fetchall(self):
         return self._fetch_all()
 
-    def fetchmany(self, size=None):
+    def fetchmany(self, size: int = None):
         # _logger.info("fetching all from overriden method")
         if size is None:
             size = self.arraysize
         if self._data is None:
             self._data = list()
         while len(self._data) < size:
-            # _logger.info("fetching next batch from fetch many")
             rows = self.fetch_batch()
             if rows is None:
                 break
@@ -530,24 +563,10 @@ class Cursor(DBAPICursor):
 
     def fetchone(self):
         # _logger.info("fetch One returning the batch itself which is limited by predefined no.of rows")
-        rows_to_return = []
-        client = self.connection.client
-        get_next_result_row_request = e6x_engine_pb2.GetNextResultRowRequest(
-            engineIP=self._engine_ip,
-            sessionId=self.connection.get_session_id,
-            queryId=self._query_id
-        )
-        get_next_result_row_response = client.getNextResultRow(get_next_result_row_request)
-        buffer = get_next_result_row_response.resultRow
-        if not self._is_metadata_updated:
-            self.update_mete_data()
-            self._is_metadata_updated = True
-        if not buffer:
+        rows = self.fetchmany(1)
+        if rows is None or len(rows) == 0:
             return None
-        buffer = BytesIO(buffer)
-        dis = DataInputStream(buffer)
-        rows_to_return.append(read_values_from_array(self._query_columns_description, dis))
-        return rows_to_return
+        return rows
 
     def explain(self):
         explain_request = e6x_engine_pb2.ExplainRequest(
@@ -555,7 +574,10 @@ class Cursor(DBAPICursor):
             sessionId=self.connection.get_session_id,
             queryId=self._query_id
         )
-        explain_response = self.connection.client.explain(explain_request)
+        explain_response = self.connection.client.explain(
+            explain_request,
+            metadata=self.metadata
+        )
         return explain_response.explain
 
     def explain_analyse(self):
@@ -564,7 +586,10 @@ class Cursor(DBAPICursor):
             sessionId=self.connection.get_session_id,
             queryId=self._query_id
         )
-        explain_analyze_response = self.connection.client.explainAnalyze(explain_analyze_request)
+        explain_analyze_response = self.connection.client.explainAnalyze(
+            explain_analyze_request,
+            metadata=self.metadata
+        )
         return dict(
             is_cached=explain_analyze_response.isCached,
             parsing_time=explain_analyze_response.parsingTime,
